package com.boman.web.core.service.save;

import com.alibaba.fastjson.JSONObject;
import com.boman.common.core.utils.SecurityUtils;
import com.boman.gen.domain.GenTableColumn;
import com.boman.web.core.constant.FormDataConstant;
import com.boman.web.core.domain.MainTableRecord;
import com.boman.web.core.domain.RowResult;
import com.boman.web.core.mapper.StandardlyMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.sql.Timestamp;
import java.util.List;

/**
 * @author shiqian
 * @description
 * @date 2021年03月22日 11:25
 **/
@Component
public class BaseSaveServiceImpl implements IBaseSaveService {

    private static final Logger LOGGER = LoggerFactory.getLogger(BaseSaveServiceImpl.class);

    @Autowired
    private StandardlyMapper mapper;

    /**
     * 功能描述: 保存一行
     *
     * @param tableName tableName
     * @param pkName    主键名称
     * @param maxId     最大的SEQ
     * @param row       要保存的数据
     * @param columns   这是此表的所有字段属性名称，用来判断是否所有UPDATE_BY和UPDATE_TIME，和时间类型
     * @return com.boman.system.common.RowResult
     */
    @Override
    public RowResult insertRow(String tableName, String pkName, Long maxId, MainTableRecord row, List<GenTableColumn> columns) {
        JSONObject commitData = row.getMainData().getCommitData();
        Timestamp currentTime = new Timestamp(System.currentTimeMillis());
        for (GenTableColumn column : columns) {
            // 判断是否有修改人、修改时间
            if (column.getColumnName().equalsIgnoreCase(FormDataConstant.UPDATE_BY)) {
                commitData.put(FormDataConstant.UPDATE_BY, SecurityUtils.getUserId());
            }
            if (column.getColumnName().equalsIgnoreCase(FormDataConstant.UPDATE_TIME)) {
                commitData.put(FormDataConstant.UPDATE_TIME, currentTime);
            }

            // 如果某一列是datetime类型，需要把类型转成datetime，否则数据库会报错
            if (FormDataConstant.DATETIME.equalsIgnoreCase(column.getColumnType())) {
                String columnName = column.getColumnName().toUpperCase();
                commitData.put(columnName, commitData.getTimestamp(columnName));
            }
        }

        // 新增
        if (row.getId() > 0) {
            commitData.put(FormDataConstant.UPDATE_TIME, currentTime);
            commitData.put(FormDataConstant.UPDATE_BY, SecurityUtils.getUserId());
        }

        commitData.put(pkName.toUpperCase(), maxId);
        int ret = mapper.insert(tableName, commitData);
        if (ret > 0) {
            commitData.put(FormDataConstant.SUCCESS_CNT, ret);
            return RowResult.ok("保存成功", commitData);
        }

        return RowResult.error("失败");
    }

}
